//-----------------------------------------------------------------------
// class INetworkGame
//
// Network game interface
//
// Author:  Sylvain Lefebvre
// History: Created 04/04/2002
//-----------------------------------------------------------------------
#ifndef __INETWORKGAME__
#define __INETWORKGAME__
//-----------------------------------------------------------------------
#include "INetworkUser.h"
#include "CNetwork.h"
//-----------------------------------------------------------------------
class IGhost;
class IGhostMaster;
//-----------------------------------------------------------------------
#define NG_MSG_NEW_GHOST           NTWRK_MSG_USER+1
#define NG_MSG_UPDATE_GHOST        NTWRK_MSG_USER+2
#define NG_MSG_REQUEST_GHOSTS      NTWRK_MSG_USER+3
#define NG_MSG_REQUEST_IDS         NTWRK_MSG_USER+4
#define NG_MSG_IDS                 NTWRK_MSG_USER+5
#define NG_MSG_GHOST               NTWRK_MSG_USER+6
#define NG_MSG_MASTER              NTWRK_MSG_USER+7
#define NG_MSG_END_OF_GHOSTS       NTWRK_MSG_USER+8
#define NG_MSG_KILL_GHOST          NTWRK_MSG_USER+9
#define NG_MSG_ALL_GHOSTS_RECEIVED NTWRK_MSG_USER+10
#define NG_MSG_USER                NTWRK_MSG_USER+128
//-----------------------------------------------------------------------
#define NG_NB_IDS_REQUESTED      1000000
#define NG_GHOST_MSG_BUFFER_SIZE 64
//-----------------------------------------------------------------------
typedef struct ng_new_ghost
{
  int            m_iId;
  int            m_iType;
  Uint8          m_iIdMaster;
  // data follows
}ng_new_ghost;
//-----------------------------------------------------------------------
typedef struct ng_update_ghost
{
  int            m_iId;
  // data follows
}ng_update_ghost;
//-----------------------------------------------------------------------
typedef struct ng_kill_ghost
{
  int            m_iId;
}ng_kill_ghost;
//-----------------------------------------------------------------------
typedef struct ng_ids
{
  int            m_iFirst;
  int            m_iLast;
}ng_ids;
//-----------------------------------------------------------------------
typedef struct ng_ghost_msg
{
  int            m_iId;
  int            m_iInfo;
  // data follows
}ng_ghost_msg;
//-----------------------------------------------------------------------
class INetworkGame : public INetworkUser
{
private:

  int                     m_iNextId;
  int                     m_iLastId;
  map<int,IGhost *>       m_Ghosts;
  map<int,IGhostMaster *> m_GhostMasters;
  bool                    m_bUpdateGhostEnabled;
  bool                    m_bIdsReserved;
  
  typedef map<int,IGhost *>::iterator       ghost_iterator;
  typedef map<int,IGhostMaster *>::iterator ghostmaster_iterator;

protected:

  // methods from INetworkUser
  virtual void    processServerMessage(const network_msg&);
  virtual void    processClientMessage(const network_msg&);
  virtual void    onNetworkReady();

  // method for ghost creation
  virtual IGhost *createGhost(int t,int id,
			      const unsigned char *d,int l)=0;

  // methods for ghost handling
  int     nextGhostId();
  void    requestIds();
  void    requestGhosts();
  void    sendGhosts(Uint8);
  void    updateGhostsOf(IGhostMaster *);
  void    prepareUpdateGhostData(IGhostMaster *,
				 unsigned char*&,int&);
  void    prepareNewGhostData(IGhostMaster *,
			      unsigned char*&,int&);
  void    prepareNewGhostData(IGhost *,
			      unsigned char*&,int&);
  void    prepareGhostMsgData(int id,int info,
			      const unsigned char *din,
			      int                  lin,
			      unsigned char*&      dout,
			      int&                 lout);

  // events
  virtual void onNewGhost(IGhost *) {}
  virtual void onAllGhostsReceived(int) {}
  virtual void onClientDisconnection(int cli);
  virtual void onClientConnection(int cli);

public:
  INetworkGame();
  virtual ~INetworkGame();
  
  void            registerGhostMaster(IGhostMaster *);
  void            unregisterGhostMaster(IGhostMaster *);
  void            removeGhost(IGhost *);

  virtual void    waitForNetworkReady();
  virtual void    processGhosts();

  IGhost       *ghostFromId(int);
  IGhostMaster *masterFromId(int);

  // messages between ghosts and master
  void ghostToMaster (IGhost *,int,const unsigned char *,int);
  void masterToGhost (IGhostMaster *,int,int,const unsigned char *,int);
  void masterToGhosts(IGhostMaster *,int,const unsigned char *,int);
};
//-----------------------------------------------------------------------
#endif
//-----------------------------------------------------------------------
// ---------------------------------------------------------- 
// Sylvain Lefebvre - (c) 2002 - Sylvain.Lefebvre@laposte.net
// ----------------------------------------------------------
