//----------------------------------------------------------------------

// TODO : loading RGBA font support

//----------------------------------------------------------------------

#include <stdio.h>

#include <string.h>

#include <iostream>

//----------------------------------------------------------------------

#ifdef WIN32

# include <windows.h>

#endif

//----------------------------------------------------------------------

#include <GL/gl.h>

//----------------------------------------------------------------------

#include "gltex.h"

#include "CFont.h"

#include "CException.h"

#include "gltex.h"

//----------------------------------------------------------------------

#ifndef MAX

#  define MAX(a,b) ((a)>(b)?(a):(b))

#endif

//----------------------------------------------------------------------

#define HAS_A_HOLE(c) ((c) == '"')

//----------------------------------------------------------------------

char *CFont::m_szBuf=NULL;

int   CFont::m_iBufLength=0;

//----------------------------------------------------------------------

//  Constructeur de base

//----------------------------------------------------------------------

CFont::CFont(const char *n,bool alpha)

{

  m_Tex=CTexture::loadTexture(n);

  if (m_Tex->isAlpha())

    throw CException("Font with alpha layer not yet supported. Sorry.");

  // init

  memset(m_CarSizeX,0,sizeof(double)*256);

  // detecte les caracteres

  calculeHauteur();

  calculeLargeur();

  // pour faciliter le boulot a OpenGL, la font est pulverisee sur une image carree

  pulverise();

  // caractere espace

  m_CarSizeX[' ']=0.5*m_dMaxCarW;

  // cree les listes de rendu

  genRenderLists();

  // alpha ?

  if (alpha)

    m_Tex->convert2Alpha(0,0,0,0);

  // create gl texture

  m_uiTexId=gltex::gltexIdFromTexture(m_Tex);

}

//----------------------------------------------------------------------

CFont::~CFont()

{

  glDeleteTextures(1,&m_uiTexId);

  delete (m_Tex);

}

//----------------------------------------------------------------------

void CFont::calculeLargeur()

{

  int  car,x,y;

  bool nothing,hole,hole_ok,in;

  unsigned char *data=m_Tex->getData();

	

  car=CFONT_FIRST_CAR;

  m_iMaxCarW=0;

  hole=false;	// premier caractere == '!'

  hole_ok=false;

  in=false;

  for (x=0;x<m_Tex->getWidth();x++)

  {

    nothing=true;

    for (y=0;y<m_Tex->getHeight();y++)

    {

      if (   data[(x+y*m_Tex->getWidth())*3  ] > 0

	     || data[(x+y*m_Tex->getWidth())*3+1] > 0

	     || data[(x+y*m_Tex->getWidth())*3+2] > 0)

      {

	if (!in)

	{

	  in=true;

	  if (!(hole && hole_ok))

	    m_CarBeginX[car]=x;

	}

	nothing=false;

	break;

      }

    }

    if (in && nothing)

    {

      in=false;

      if (hole && !hole_ok)

      {

	hole_ok=true;

      }

      else

      {

	m_CarEndX[car]=x;

	m_CarSizeX[car]=m_CarEndX[car]-m_CarBeginX[car]+1;

	if (m_CarSizeX[car] > m_iMaxCarW)

	  m_iMaxCarW=(int)m_CarSizeX[car];

	car++;

	if (HAS_A_HOLE(car))

	{

	  hole=true;

	  hole_ok=false;

	}

	else

	{

	  hole=false;

	}

	if (car > 255)

	  throw CException("Trop de caracteres dans la font %s !",m_Tex->getName());

      }

    }

  }

  if (car-CFONT_FIRST_CAR < CFONT_NB_CAR)

    throw CException("Pas assez de caracteres dans la font %s (%d/%d) !",m_Tex->getName(),car-CFONT_FIRST_CAR,CFONT_NB_CAR);

  if (car-CFONT_FIRST_CAR > CFONT_NB_CAR)

    throw CException("Trop de caracteres dans la font %s (%d/%d) !",m_Tex->getName(),car-CFONT_FIRST_CAR,CFONT_NB_CAR);

}

//----------------------------------------------------------------------

void CFont::calculeHauteur()

{

  int  x,y;

  unsigned char *data=m_Tex->getData();

	

  // determine la hauteur min et max

  m_iCarMinY=0;

  m_iCarMaxY=m_Tex->getHeight();

  for (y=0;y<m_Tex->getHeight();y++)

  {

    for (x=0;x<m_Tex->getWidth();x++)

    {

      if (   data[(x+y*m_Tex->getWidth())*3  ] > 0

	     || data[(x+y*m_Tex->getWidth())*3+1] > 0

	     || data[(x+y*m_Tex->getWidth())*3+2] > 0)

      {

	if (x < m_iCarMinY)

	  m_iCarMinY=y;

	if (x > m_iCarMaxY)

	  m_iCarMaxY=y;

      }

    }

  }

}

//----------------------------------------------------------------------

void CFont::pulverise()

{

  unsigned int   n;

  unsigned char *ndata;

  char		 car;

	

  // grille 10 par 10 de caracteres

  n=MAX(m_iMaxCarW,(m_iCarMaxY-m_iCarMinY+1))*10;

  ndata=new unsigned char[n*n*3];

  // tout a zero

  memset(ndata,0,n*n*3);

  // transfert des caracteres

  for (car=CFONT_FIRST_CAR;car<=CFONT_LAST_CAR;car++)

    transfertCar(car,n,ndata);

  // create new texture

  delete (m_Tex);

  m_Tex=new CTexture("font_internal",n,n,false,ndata);

  m_Tex->setDataOwner(true);

  // la taille est exprimee dans l'espace texture normalise

  for (car=CFONT_FIRST_CAR;car<=CFONT_LAST_CAR;car++)

    m_CarSizeX[car]/=((double)m_Tex->getWidth());

  m_dCarMinY=((double)m_iCarMinY)/((double)m_Tex->getHeight());

  m_dCarMaxY=((double)m_iCarMaxY)/((double)m_Tex->getHeight());

  m_dMaxCarW=((double)m_iMaxCarW)/((double)m_Tex->getWidth());

}

//----------------------------------------------------------------------

void CFont::transfertCar(int car,int n,unsigned char *ndata)

{

  int i,j,ki,kj,u,v,cy;

  unsigned char *data=m_Tex->getData();

	

  cy=car-CFONT_FIRST_CAR;

  u=((cy % 10)*n)/10;

  v=((cy / 10)*n)/10;

  for (j=(int)m_iCarMinY;j<(int)m_iCarMaxY;j++)

  {

    kj=j-(int)m_iCarMinY;

    for (i=(int)m_CarBeginX[car];i<(int)m_CarEndX[car];i++)

    {

      ki=i-(int)m_CarBeginX[car];

      ndata[((u+ki+(v+kj)*n)*3  )%(n*n*3)]=data[(i+j*m_Tex->getWidth())*3  ];

      ndata[((u+ki+(v+kj)*n)*3+1)%(n*n*3)]=data[(i+j*m_Tex->getWidth())*3+1];

      ndata[((u+ki+(v+kj)*n)*3+2)%(n*n*3)]=data[(i+j*m_Tex->getWidth())*3+2];

    }

  }

}

//----------------------------------------------------------------------

void CFont::genRenderLists()

{

  int    car,cy;

  double u,v,larg;

	

  m_uiCarRenderLists=glGenLists(CFONT_NB_CAR+1);

  for (car=CFONT_FIRST_CAR-1;car<=CFONT_LAST_CAR;car++) // commence sur l'espace FIXME pas tres joli ce CFONT_FIRST_CAR-1

  {

    glNewList(m_uiCarRenderLists+car-(CFONT_FIRST_CAR-1),GL_COMPILE);

    if (car == CFONT_FIRST_CAR-1) // (car == ' ')

    {

      glTranslated(0.5,0.0,0.0);

    }

    else

    {

      // debut caractere

      larg=m_CarSizeX[car]/m_dMaxCarW;

      cy=car-CFONT_FIRST_CAR;

      u=((double)(cy % 10))/10.0;

      v=((double)(cy / 10))/10.0;

      glBegin(GL_QUADS);

      // p1

      glTexCoord2d(u,v+m_dCarMinY);

      glVertex2d(0.0,0.0);

      // p2

      glTexCoord2d(u,v+m_dCarMaxY);

      glVertex2d(0.0,1.0);

      // p3

      glTexCoord2d(u+m_CarSizeX[car],v+m_dCarMaxY);

      glVertex2d(larg,1.0);

      // p4

      glTexCoord2d(u+m_CarSizeX[car],v+m_dCarMinY);

      glVertex2d(larg,0.0);

      glEnd();

      // avance

      glTranslated(larg,0.0,0.0);

    }

    // fin du caractere

    glEndList();

  }

}

//----------------------------------------------------------------------

void CFont::printString(double x,double y,double t,const char *s)

{

  printStringN(x,y,t,s,-1);

}

//----------------------------------------------------------------------

void  CFont::copyToBuffer(const char *s)

{

  int l;

	

  // alloc dynamique du buffer d'affichage

  if (m_szBuf == NULL)

  {

    m_iBufLength=strlen(s);

    m_szBuf=new char[m_iBufLength+1];

  }

  else if (m_iBufLength < (l=strlen(s)))

  {

    m_iBufLength=l;

    delete [](m_szBuf);

    m_szBuf=new char[m_iBufLength+1];

  }

  strcpy(m_szBuf,s);

}

//----------------------------------------------------------------------

void CFont::printStringN(double x,double y,double t,const char *s,int nb)

{

  char        *subs;

  int	       l,ntoprint,nprinted=0;

	

  copyToBuffer(s);

  glPushAttrib(GL_ENABLE_BIT);

  glEnable(GL_TEXTURE_2D);

  glBindTexture(GL_TEXTURE_2D,m_uiTexId);

  if (m_Tex->isAlpha())

  {

    glEnable(GL_ALPHA_TEST);

    glAlphaFunc(GL_GREATER,0.0);

  }

  else

    glDisable(GL_ALPHA_TEST);

  glListBase(m_uiCarRenderLists-(CFONT_FIRST_CAR-1));

  glPushMatrix();

  glTranslated(x,y,0.0);

  subs=strtok(m_szBuf,"\n");

  if (subs != NULL)

  {

    do

    {

      glPushMatrix();

      glScaled(t,t,1.0);

      l=strlen(subs);

      if (nb >= 0)

      {

	if (l+nprinted > nb)

	  ntoprint=nb-nprinted;

      }

      else

	ntoprint=l;

      glCallLists(ntoprint,GL_UNSIGNED_BYTE,subs);

      nprinted+=ntoprint;

      glPopMatrix();

      if (nb >= 0 && nprinted >= nb)

	break;

      glTranslated(0.0,t,0.0);

    }

    while ((subs=strtok(NULL,"\n")) != NULL);

  }

  glListBase(0);

  glPopMatrix();

  glPopAttrib();

}

//----------------------------------------------------------------------

void CFont::printStringNeed(double t,const char *s,double *w,double *h)

{

  char  *subs;

  int    i,l;

  double wtmp;

	

  copyToBuffer(s);

  *h=0.0;

  *w=0.0;

  subs=strtok(m_szBuf,"\n");

  if (subs != NULL)

  {

    do

    {

      *h+=t;

      l=strlen(subs);

      wtmp=0.0;

      for (i=0;i<l;i++)

	wtmp+=t*m_CarSizeX[subs[i]]/m_dMaxCarW;

      if (wtmp > *w)

	*w=wtmp;

    }

    while ((subs=strtok(NULL,"\n")) != NULL);

  }

}

//----------------------------------------------------------------------

