header "pre_include_hpp" 
{

#include <cmath>
#include <map>
#include <string>

class CMobile;

#include "atr_nfo.h"

}

header "pre_include_cpp" 
{
#include "CException.h"
#include "CMobile.h"
#include "CWorld.h"

#define ME             -1
#define MOUSE          -2
#define COLLIDER       -3
#define PREV           -4
#define NEXT           -5
#define NEAREST        -6
#define NEAREST2       -7
#define NEAREST3       -8
#define NEAREST_LEFT   -9
#define NEAREST_RIGHT -10
#define NEAREST_UP    -11
#define NEAREST_DOWN  -12
#define EMPTY         -13
#define NEAREST_INDIR -14
}


options 
{
  language="Cpp";
}

/*  **************************************************************
    LE PARSEUR D'AST (execute les noeuds une fois analyse's)
*/

{
  // nfo <- identification de l'attribut nom du mobile i du client j
    void CalcWalker::set_atr_nfo(atr_nfo &nfo, int i, int j, string nom, int var)
    {
        if( -1 != j )
            throw CException("Attempt to modify another guy !");

        if ( -1 == i)
            nfo.m = m_Mobile;
        else
            nfo.m=WORLD.getLocalMobile(i - 1);

        nfo.name = nom;
        nfo.variable = var;
    }

    // retourne un pointeur sur un mobile (i, i@j, ou special)
    CMobile *CalcWalker::getMobilePtr(mob_adress mob)
    {
        CMobile *m=NULL;

        int ia = mob.mobile;
        int ib = mob.client;

        if( ia == -100)
            return NULL;

        if( ia < -1 )
        {
            switch(ia)
            {
            case MOUSE :         // mouse
                m = WORLD.getMouse(); break;
            case COLLIDER :      // collider
                m = m_Collider;       break;
            case PREV :          // get previous mobile
                m=WORLD.getPrevMobile(m_Mobile);  break;
            case NEXT :          // get next mobile
                m=WORLD.getNextMobile(m_Mobile);  break;
            case NEAREST :       // plusproche
                m = WORLD.getNearest(m_Mobile);   break;
            case NEAREST2 :      // plusproche
                m = WORLD.getNearest(m_Mobile,2); break;
            case NEAREST3 :      // plusproche
                m = WORLD.getNearest(m_Mobile,3); break;
            case NEAREST_LEFT :  // plusproche_gauche
                m = WORLD.getNearestDir(m_Mobile,-1,0,M_PI/2); break;
            case NEAREST_RIGHT : // plusproche_droit
                m = WORLD.getNearestDir(m_Mobile,1,0,M_PI/2);  break;
            case NEAREST_UP :    // plusproche_haut
                m = WORLD.getNearestDir(m_Mobile,0,1,M_PI/2);  break;
            case NEAREST_DOWN :  // plusproche_bas
                m = WORLD.getNearestDir(m_Mobile,0,-1,M_PI/2); break;
	    case EMPTY:          // NULL
	        m = NULL;            break;
	    case NEAREST_INDIR:
		m = WORLD.getNearestDir(m_Mobile, mob.dx, mob.dy, mob.angle);
		break;
            default: break;
            }
        } 
        else if (ib > -1)
        {
            // mobile on distant client
            if (ia == -1)
                throw CException("[ERROR] Walker - mobile id error.");
            m=WORLD.getDistantMobile(ia-1,ib-1);
        }
        else
        {
            // mobile on local client
            if (ia > -1)
            {
                m=WORLD.getLocalMobile(ia-1);
            }
            else if( ia == -1 )
            {
                m=m_Mobile;
            }
        }
        
        return m;
    }

    // retourne la valeur de l'attribut attrib (sauf mem) du mobile mob
    double CalcWalker::get_atr_nfo(int attrib, mob_adress mob, int var)
    {
        double r;
        CMobile *m= getMobilePtr(mob);
        // give its value to the attribute
        r = 0.0;
        if( m )
        {
            switch( attrib )
            {
            case ATTR_X:  r=m->pos().x(); break;
            case ATTR_Y:  r=m->pos().y(); break;
            case ATTR_DX: r=m->speed().x(); break;
            case ATTR_DY: r=m->speed().y(); break;
            case ATTR_ANGLE: r=m->getAngle(); break;
            case ATTR_ICONE: r=m->getIcon(); break;
            case ATTR_VAR: r=m->getVar(var); break;
            case ATTR_COULEUR: r=m->getColor(); break;
            case ATTR_ROUGE: r= ( ( m->getColor() & 0xff000000 ) >> 24 )/255.; break;
            case ATTR_VERT:  r= ( ( m->getColor() & 0x00ff0000 ) >> 16 )/255.; break;
            case ATTR_BLEU:  r= ( ( m->getColor() & 0x0000ff00 ) >>  8 )/255.; break;
            case ATTR_GRIS:  {
				r  =  ( ( m->getColor() & 0x0000ff00 ) >>  8);
				r +=  ( ( m->getColor() & 0x00ff0000 ) >> 16);
				r +=  ( ( m->getColor() & 0xff000000 ) >> 24);
				r /= (3.0*255.);
				break;
			     }
            case ATTR_LARGEUR: r=m->getWidth(); break;
            case ATTR_HAUTEUR: r=m->getHeight(); break;
            case ATTR_CHOCX: r=m_Choc.x(); break;
            case ATTR_CHOCY: r=m_Choc.y(); break;
            case ATTR_VISIBLE: r= ( m->isVisible() ? 1.0 : 0.0 ); break;
            case ATTR_CONTACT: r= ( m->firstContact() ? 1.0 : 0.0 ); break;

            default: r = 0.0; break;
            }
        }
        return r;
    }

    // retourne un pointeur sur une variable temporaire float
    double *CalcWalker::getFloatRef(string s, bool createOnFail)
    {
        fltMap_t::iterator it = mFloats.find(s);
        if( it != mFloats.end() )
        {
            return &((*it).second);
        }
        else if( createOnFail )
        {
            mFloats[s]= 0.0;
            return &(mFloats[s]);
        }
        else
            return NULL;
    }

    // retourne un pointeur sur une variable temporaire ref
    mob_adress *CalcWalker::getMobileRef(string s, bool createOnFail)
    {
        mobMap_t::iterator it = mMobiles.find(s);
        if( it != mMobiles.end() )
        {
            return &((*it).second);
        }
        else if( createOnFail )
        {
            mob_adress a;
            a.mobile = -1;
            a.client = -1;
            mMobiles[s]=a;
            return  &(mMobiles[s]);
        }
        else
            return NULL;
    }

    void CalcWalker::push_state()
    {
//        cerr << "PUSH STATE" << endl;
        snapshot ns;
        ns.mFloats = mFloats;
        ns.mMobiles = mMobiles;
        ns.mMobileData = *((const mobile_data *)m_Mobile->writeNetworkData());
        mMemory.push_front(ns);
    }
    void CalcWalker::pop_state()
    {
//        cerr << "POP STATE" << endl;
        mMemory.pop_front();
    }
    void CalcWalker::restore_state()
    {
//        cerr << "RESTORE STATE size = " << mMemory.size() << endl;
        snapshot ns = mMemory.front();
        mMemory.pop_front();
        mFloats = ns.mFloats;
        mMobiles = ns.mMobiles;
        m_Mobile->readNetworkData((const unsigned char *)&(ns.mMobileData), 0);
    }
}
class CalcWalker extends TreeParser;
{
private:
    CMobile *m_Mobile, *m_Collider;
    CVertex m_Choc;
    int doNotInterprete;
    typedef std::map<string, double> fltMap_t;
    fltMap_t mFloats;
    typedef std::map<string, mob_adress> mobMap_t;
    mobMap_t mMobiles;

    class snapshot
    {
        public:
        fltMap_t mFloats;
        mobMap_t mMobiles;
        mobile_data mMobileData;
        public:
        snapshot() : mFloats(), mMobiles() {}
    };
    list<snapshot> mMemory;

public:
    void bindTo(CMobile *m, CMobile *collider=NULL, CVertex choc = CVertex(0.0, 0.0, 0.0))
    {
        m_Mobile=m;
        m_Collider=collider;
        mFloats.clear();
        mMobiles.clear();
        mMemory.clear();
        doNotInterprete = 0;
        m_Choc = choc;
    }
    void push_state();
    void pop_state();
    void restore_state();
    double get_atr_nfo(int, mob_adress, int);
    void set_atr_nfo(atr_nfo&, int, int, string, int);
    CMobile *getMobilePtr(mob_adress);
    double *getFloatRef(string, bool);
    mob_adress *getMobileRef(string, bool);
}

start
    : ( expr )*
    ;

// ----- execution des commandes ------------------
expr
{
    double    a, b, cond;
    atr_nfo  atr;
    mob_adress mob;
}
    : #("if" cond=F:fexpr)
        {
            if( cond != 0.0 )
	    {
                if( F->getNextSibling() )
		    if( F->getNextSibling()->getFirstChild() )
			start(F->getNextSibling()->getFirstChild());
	    }
	    else
	    {
                if( F->getNextSibling() )
		{
		    if( F->getNextSibling()->getNextSibling() )
		    {
			start( F->getNextSibling()->getNextSibling() );
		    }
		}
	    }
        }
    | #("stop" mob=ref)
        {
            CMobile * mref = NULL;
            mref = getMobilePtr(mob);
            //if( doNotInterprete )
              //  cerr << "stop je fais rien!" << endl;
            if( mref && !doNotInterprete )
                mref->execStop();
        }
    | #("start" mob=ref)
        {
            CMobile * mref = NULL;
            mref = getMobilePtr(mob);
            //if( doNotInterprete )
              //  cerr << "start je fais rien!" << endl;
            if( mref && !doNotInterprete )
                mref->execStart();
        }
    | #("restart" mob=ref)
        {
            CMobile * mref = NULL;
            mref = getMobilePtr(mob);
           // if( doNotInterprete )
            //    cerr << "restart je fais rien!" << endl;
            if( mref && !doNotInterprete )
            {
                mref->execStop();
                mref->execStart();
            }
        }
    | #("move_to" mob=ref a=fexpr b=fexpr)
        {
            CMobile * mref = NULL;
            mref = getMobilePtr(mob);
            if( mref )
                mref->execMoveTo(a, b);
        }

// ----- affectation d'attributs ------------------

    | (#(COLON ID ref))=> #(COLON id:ID mob=ref)
        {
            mob_adress * mref = NULL;
            mref = getMobileRef(id->getText(), true);
            if( mref && !doNotInterprete )
                *mref = mob;
        }
    | (#(COLON ID fexpr))=> #(COLON id2:ID a=fexpr)
        {
            double * fref = NULL;
            fref = getFloatRef(id2->getText(), true);
            if( fref && !doNotInterprete )
                *fref = a;
        }
    | (#(COLON attribref fexpr))=> #(COLON atr=attribref a=fexpr)
        {
            if( !doNotInterprete )
            {
            if (atr.name == "x") {
                atr.m->pos().setX(a);
            }
            else if (atr.name == "y") {
                atr.m->pos().setY(a);
            }
            else if (atr.name == "dx") {
                atr.m->speed().setX(a);
            }
            else if (atr.name == "dy") {
                atr.m->speed().setY(a);
            }
            else if (atr.name == "angle") {
                atr.m->setAngle(a);
            }
            else if (atr.name == "icon") {
                atr.m->setIcon((int) a);
            }
            else if (atr.name == "mem") {
                atr.m->setVar(atr.variable, a);
            }
            else if (atr.name == "color") {
                atr.m->setColor((unsigned int)a);
            }
	    else if (atr.name == "red") {
                unsigned int c;
                c = (unsigned int)( (a<=0)?0: (a>=1.)? 255.: 255.*a);
		atr.m->setColor( (atr.m->getColor() & 0x00ffff00) | ( c << 24 ) );
	    }
            else if (atr.name == "green") {
                unsigned int c;
                c = (unsigned int)( (a<=0)?0: (a>=1.)? 255.: 255.*a);
                atr.m->setColor( (atr.m->getColor() & 0xff00ff00) | ( c << 16 ) );
            }
            else if (atr.name == "blue") {
                unsigned int c;
                c = (unsigned int)( (a<=0)?0: (a>=1.)? 255.: 255.*a);
                atr.m->setColor( (atr.m->getColor() & 0xffff0000) | ( c <<  8 ) );
            }
            else if (atr.name == "gray") {
                unsigned int c;
                c = (unsigned int)( (a<=0)?0: (a>=1.)? 255.: 255.*a);
                atr.m->setColor( (atr.m->getColor() & 0x00ffff00) | ( c << 24 ) );
                atr.m->setColor( (atr.m->getColor() & 0xff00ff00) | ( c << 16 ) );
                atr.m->setColor( (atr.m->getColor() & 0xffff0000) | ( c <<  8 ) );
            }
            else if (atr.name == "width") {
                atr.m->setWidth(a);
            }
            else if (atr.name == "height") {
                atr.m->setHeight(a);
            }
            else if (atr.name == "visible") {
                atr.m->setVisible(a != 0.0);
		atr.m->updateVisible();
            }
            else if (atr.name == "contact") {
                    // nothing to do.
            }
            else
                throw CException("[ERROR] Walker - Unknown attribute !");
            }
        }
    ;

ref returns [mob_adress mob]
{
    double i,j,dy,ang;
}
    : #(REF i=fexpr j=fexpr dy=fexpr ang=fexpr)
        {
            mob.mobile = (int)i;
            mob.client = (int)j;
	    mob.dx = j;
	    mob.dy = dy;
	    mob.angle = ang;
        }
    | id:ID
        {
            mob_adress *mref = NULL;
            mref = getMobileRef(id->getText(), false);
            if( mref )
                mob = *mref;
            else
	    {
	        double *floatref = NULL;
		floatref = getFloatRef(id->getText(), false);
		if( floatref != NULL )
		{
                	mob.mobile = ((int)(*floatref));
			mob.client = -1;
		}
            	else
            	{
                	mob.mobile = -100;
               		mob.client = -100;
           	}
	    }
        }
    ;

// an attrib as a left value
attribref returns [atr_nfo nfo]
{
    mob_adress mob;
}
    : #("x" mob=ref) { set_atr_nfo(nfo, mob.mobile, mob.client, "x", 0); }
    | #("y" mob=ref) { set_atr_nfo(nfo, mob.mobile, mob.client, "y", 0); }
    | #("dx" mob=ref) { set_atr_nfo(nfo, mob.mobile, mob.client, "dx", 0); }
    | #("dy" mob=ref) { set_atr_nfo(nfo, mob.mobile, mob.client, "dy", 0); }
    | #("angle" mob=ref) { set_atr_nfo(nfo, mob.mobile, mob.client, "angle", 0); }
    | #("icon" mob=ref) { set_atr_nfo(nfo, mob.mobile, mob.client, "icon", 0); }
    | #("color" mob=ref) { set_atr_nfo(nfo, mob.mobile, mob.client, "color", 0); }
    | #("red" mob=ref) { set_atr_nfo(nfo, mob.mobile, mob.client, "red", 0); }
    | #("green" mob=ref) { set_atr_nfo(nfo, mob.mobile, mob.client, "green", 0); }
    | #("blue" mob=ref) { set_atr_nfo(nfo, mob.mobile, mob.client, "blue", 0); }
    | #("gray" mob=ref) { set_atr_nfo(nfo, mob.mobile, mob.client, "gray", 0); }
    | #("shockx" mob=ref) { set_atr_nfo(nfo, mob.mobile, mob.client, "shockx", 0); }
    | #("shocky" mob=ref) { set_atr_nfo(nfo, mob.mobile, mob.client, "shocky", 0); }
    | #("width" mob=ref) { set_atr_nfo(nfo, mob.mobile, mob.client, "width", 0); }
    | #("height" mob=ref) { set_atr_nfo(nfo, mob.mobile, mob.client, "height", 0); }
    | #("visible" mob=ref) { set_atr_nfo(nfo, mob.mobile, mob.client, "visible", 0); }
    | #("contact" mob=ref) { set_atr_nfo(nfo, mob.mobile, mob.client, "contact", 0); }
    | #(i:"mem" mob=ref) { set_atr_nfo(nfo, mob.mobile, mob.client, "mem", atof(i->getText().c_str())); }
    ;

// ----- fonctions a evaluer ------------------
// "right" values
fexpr returns [double r]
{
    double a, b;
    mob_adress mob, mob2;
}
    : #(PLUS a=fexpr b=fexpr) { r = a + b; }
    | (#(MINUS fexpr fexpr))=> #(MINUS a=fexpr b=fexpr) { r =  a - b; }
    | #(MINUS a=fexpr) { r = -a; }
    | #(STAR a=fexpr b=fexpr) { r = a * b; }
    | #(DIV a=fexpr b=fexpr) { r = a / b; }
    | f:FLOAT { r = atof(f->getText().c_str()); }
    | i:INT { r = atof(i->getText().c_str()); }
    | #("sqr" a=fexpr) { r = a * a; }
    | #("sqrt" a=fexpr) { r = sqrt(fabs(a)); }
    | #("abs" a=fexpr) { r = fabs(a); }
    | #("sin" a=fexpr) { r = sin(a); }
    | #("tan" a=fexpr) { r = tan(a); }
    | #("cos" a=fexpr) { r = cos(a); }
    | "click" { r = WORLD.getClic(); }
    | "key_left"  { r = WORLD.isKeyPressed(KEY_LEFT); }
    | "key_right" { r = WORLD.isKeyPressed(KEY_RIGHT); }
    | "key_up"    { r = WORLD.isKeyPressed(KEY_UP); }
    | "key_down"  { r = WORLD.isKeyPressed(KEY_DOWN); }
    | "key_space" { r = WORLD.isKeyPressed(KEY_SPACE); }
    | "rnd" { r = drand48(); }
    | "srnd" { r = 2.0 * drand48() - 1.0; }
    | #("ang" a=fexpr b=fexpr) { r = atan2(b,a); }
    | (#(EQUAL fexpr fexpr))=> #(EQUAL a=fexpr b=fexpr) { if( a == b ) r = 1.0; else r = 0.0; }
     | (#(EQUAL ref ref))=> #(EQUAL mob=ref mob2=ref)
      {
          if( getMobilePtr(mob) == getMobilePtr(mob2) ) r = 1.0; else r = 0.0;
      }
    | #(GREATER a=fexpr b=fexpr) { if( a > b ) r = 1.0; else r = 0.0; }
    | #(GEQ a=fexpr b=fexpr) { if( a >= b ) r = 1.0; else r = 0.0; }
    | #(LESS a=fexpr b=fexpr) { if( a < b ) r = 1.0; else r = 0.0; }
    | #(LEQ a=fexpr b=fexpr) { if( a <= b ) r = 1.0; else r = 0.0; }
    | #("and" a=fexpr b=fexpr) { if( ( a != 0.0 ) && ( b != 0.0 ) ) r = 1.0; else r = 0.0; }
    | #("or" a=fexpr b=fexpr) { if( ( a != 0.0 ) || ( b != 0.0 ) ) r = 1.0; else r = 0.0; }
    | (#(NEQ fexpr fexpr))=> #(NEQ a=fexpr b=fexpr) { if( a != b ) r = 1.0; else r = 0.0; }
    | (#(NEQ ref ref))=> #(NEQ mob=ref mob2=ref)
      {
          if( getMobilePtr(mob) != getMobilePtr(mob2) ) r = 1.0; else r = 0.0;
      }
    | #("min" a=fexpr b=fexpr) { r = a < b ? a : b; }
    | #("max" a=fexpr b=fexpr) { r = a > b ? a : b; }
    | #("int" a=fexpr) { r = floor(a); }
    | #("frac" a=fexpr){ r = a-floor(a); }
    | #("sgn" a=fexpr) { r = (a<0.) ? -1. : (a>0.)? 1. : 0.; }
    | #("norm" a=fexpr b=fexpr) { r = sqrt(a*a+b*b); }
    | #("mod" a=fexpr b=fexpr)
        {
            int aa = (int)a;
            int bb = (int)b;
            if( bb == 0 )
                r = aa;
            else
                r = aa - floor((double)aa/(double)bb) * bb;
        }
    | "t"  { r = WORLD.getTime(); } 
    | "dt" { r = WORLD.getTimeInterval(); } 
    |  #("x" mob=ref) { r = get_atr_nfo(ATTR_X, mob, 0); }
    |  #("y" mob=ref) { r = get_atr_nfo(ATTR_Y, mob, 0); }
    |  #("dx" mob=ref) { r = get_atr_nfo(ATTR_DX, mob, 0); }
    |  #("dy" mob=ref) { r = get_atr_nfo(ATTR_DY, mob, 0); }
    |  #("angle" mob=ref) { r = get_atr_nfo(ATTR_ANGLE, mob, 0); }
    |  #("icon" mob=ref) { r = get_atr_nfo(ATTR_ICONE, mob, 0); }
    |  #("color" mob=ref) { r = get_atr_nfo(ATTR_COULEUR, mob, 0); }
    |  #("red" mob=ref) { r = get_atr_nfo(ATTR_ROUGE, mob, 0); }
    |  #("green" mob=ref) { r = get_atr_nfo(ATTR_VERT, mob, 0); }
    |  #("blue" mob=ref) { r = get_atr_nfo(ATTR_BLEU, mob, 0); }
    |  #("gray" mob=ref) { r = get_atr_nfo(ATTR_GRIS, mob, 0); }
    |  #("shockx" mob=ref) { r = get_atr_nfo(ATTR_CHOCX, mob, 0); }
    |  #("shocky" mob=ref) { r = get_atr_nfo(ATTR_CHOCY, mob, 0); }
    |  #("width" mob=ref) { r = get_atr_nfo(ATTR_LARGEUR, mob, 0); }
    |  #("height" mob=ref) { r = get_atr_nfo(ATTR_HAUTEUR, mob, 0); }
    |  #("visible" mob=ref) { r = get_atr_nfo(ATTR_VISIBLE, mob, 0); }
    |  #("contact" mob=ref) { r = get_atr_nfo(ATTR_CONTACT, mob, 0); }
    | #(va:"mem" mob=ref) { r = get_atr_nfo(ATTR_VAR, mob, atoi(va->getText().c_str())); }
    | #("dist" mob=ref mob2=ref)
        {
            CMobile * mref = NULL;
            mref = getMobilePtr(mob);
            CMobile * mref2 = NULL;
            mref2 = getMobilePtr(mob2);
            if( ( mref != NULL ) && ( mref2 != NULL ) )
                r = mref->getPos().dist(mref2->getPos());
            else
                r = 0.0;
        }
    |  id:ID
        {
            double * fref = NULL;
            fref = getFloatRef(id->getText(), false);
            if( fref )
                r = *fref;
            else
                r = 0.0;
        }
    ;











/*  **************************************************************
    LE PARSER : construct the AST tree (Abstract Syntax Tree)
*/




class CalcParser extends Parser;
options {
	buildAST = true;
	defaultErrorHandler=false;
}

start
    : ( (expr)? ( NEWLINE! | SEMI! ) )*
    ;

isolated_start
    : a:start
	{ #isolated_start = #([REF, "BIDON"], a); }
    ;
    
expr
    : (ID COLON^ fullref)=> ID COLON^ fullref
    | lvalue COLON^ fexpr
    | "if"^ fexpr "then"! isolated_start ( "else"! start )? "endif"!
    | "stop"^  ref
    | "start"^  ref
    | "restart"^ ref
    | "move_to"^ LPAREN! ref COMMA! fexpr COMMA! fexpr RPAREN!
    ;

lvalue
    :! va2:VAR
        { #lvalue = #(va2, #([REF, "REF"], [INT,"-1"], [INT,"-1"])); }
    |! (attribute localref)=> (a:attribute l:localref)
        { #lvalue = #(a, l); }
    |! b:attribute
        { #lvalue = #(b, #([REF, "REF"], [INT,"-1"], [INT,"-1"])); }
    | ID
    ;

attribute
    : "dx" | "dy" | "x" | "y" | "angle" | "icon" | "width" | "height" | "color" | "red" | "green" | "blue" | "gray" | "shockx" | "shocky" | "visible" | "contact"
    |! "mem" i:INT
        { #attribute = #([LITERAL_mem, i->getText()]); }
    ;

localref
    : i1:INT
        { #localref = #([REF, "REF"], i1, [INT,"-1"]); }
    | ID
    ;

fullref
    :! ( INT AT )=> ( i:INT  AT j:INT )
        { #fullref = #([REF, "REF"], i, j); }
    |! "me"
        { #fullref = #([REF, "REF"], [INT,"-1"], [INT,"-1"], [INT,"-1"]); }
    |! "mouse"
        { #fullref = #([REF, "REF"], [INT,"-2"], [INT,"-1"], [INT,"-1"]); }
    |! "c"
        { #fullref = #([REF, "REF"], [INT,"-3"], [INT,"-1"], [INT,"-1"]); }
    |! "prev"
        { #fullref = #([REF, "REF"], [INT,"-4"], [INT,"-1"], [INT,"-1"]); }
    |! "next"
        { #fullref = #([REF, "REF"], [INT,"-5"], [INT,"-1"], [INT,"-1"]); }
    |! "nearest"
        { #fullref = #([REF, "REF"], [INT,"-6"], [INT,"-1"], [INT,"-1"]); }
    |! "nearest_second"
        { #fullref = #([REF, "REF"], [INT,"-7"], [INT,"-1"], [INT,"-1"]); }
    |! "nearest_third"
        { #fullref = #([REF, "REF"], [INT,"-8"], [INT,"-1"], [INT,"-1"]); }
    |! "nearest_left"
        { #fullref = #([REF, "REF"], [INT,"-9"], [INT,"-1"], [INT,"-1"]); }
    |! "nearest_right"
        { #fullref = #([REF, "REF"], [INT,"-10"], [INT,"-1"], [INT,"-1"]); }
    |! "nearest_up"
        { #fullref = #([REF, "REF"], [INT,"-11"], [INT,"-1"], [INT,"-1"]); }
    |! "nearest_down"
        { #fullref = #([REF, "REF"], [INT,"-12"], [INT,"-1"], [INT,"-1"]); }
    |! "EMPTY"
        { #fullref = #([REF, "REF"], [INT,"-13"], [INT,"-1"], [INT,"-1"]);  }
    |! "nearest_in_dir" LPAREN dx:fexpr COMMA dy:fexpr COMMA ang:fexpr RPAREN
        { #fullref = #([REF, "REF"], [INT,"-14"], dx, dy, ang); }
//    | ID
    ;

ref
    :! (fullref)=> l:fullref
        { #ref = #l; }
    |! ll:localref
        { #ref = #ll; }
    ;

fexpr : logicexpr ( ( "or"^ | "and"^ ) logicexpr )*
	 ;

logicexpr :  (ref (EQUAL | NEQ) ref) => ref ( EQUAL^ | NEQ^ ) ref
      | addexpr ( ( EQUAL^ | GREATER^ | GEQ^ | LESS^ | LEQ^ | NEQ^ ) addexpr )*
	 ;

addexpr : mexpr ( ( PLUS^ | MINUS^ ) mexpr )*
    ;

mexpr : atom ( ( STAR^ | DIV^ | "mod"^ ) atom )*
    ;

//fexpr : mexpr ( ( PLUS^ | MINUS^ ) mexpr )*
//    ;
//
//mexpr : testexpr ( ( STAR^ | DIV^ | "mod"^ ) testexpr )*
//    ;
//
////testexpr :  (fullref) => fullref ( EQUAL^ | NEQ^ ) fullref
//testexpr :  (fullref) => fullref ( EQUAL^ | NEQ^ ) ref
//         | atom ( ( "or"^ | "and"^ | EQUAL^ | GREATER^ | GEQ^ | LESS^ | LEQ^ | NEQ^ ) atom )?
//    ;

atom
    : ( PLUS^ | MINUS^ ) atom
    | float_
    |! (attribute ref)=> a:attribute r:ref { #atom = #(a, r); }
    |! a1:attribute { #atom = #(a1, #([REF, "REF"], [INT,"-1"], [INT,"-1"])); }
    | LPAREN! fexpr RPAREN!
    |! ("PI"|"pi"|"Pi") { #atom =#([FLOAT, "3.14159"]); }
    |! "RED" { #atom =#([FLOAT, "0xff000000"]); }
    |! "GREEN" { #atom =#([FLOAT, "0x00ff0000"]); }
    |! "BLUE" { #atom =#([FLOAT, "0x0000ff00"]); }
    |! "YELLOW" { #atom =#([FLOAT, "0xffff0000"]); }
    |! "CYAN" { #atom =#([FLOAT, "0x00ffff00"]); }
    |! "PURPLE" { #atom =#([FLOAT, "0xff00ff00"]); }
    |! "WHITE" { #atom =#([FLOAT, "0xffffff00"]); }
    |! "BLACK" { #atom =#([FLOAT, "0x00000000"]); }
    |! "ORANGE" { #atom =#([FLOAT, "0xff880000"]); }
    | "sqrt"^ LPAREN! fexpr RPAREN!
    | "sqr"^ LPAREN! fexpr RPAREN!
    | "abs"^ LPAREN! fexpr RPAREN!
    | "sin"^ LPAREN! fexpr RPAREN!
    | "cos"^ LPAREN! fexpr RPAREN!
    | "tan"^ LPAREN! fexpr RPAREN!
    | "rnd"^
    | "srnd"^
    | "click"^
    | "key_left"^
    | "key_right"^
    | "key_up"^
    | "key_down"^
    | "key_space"^
    | "ang"^ LPAREN! fexpr COMMA! fexpr RPAREN!
    | "min"^ LPAREN! fexpr COMMA! fexpr RPAREN!
    | "max"^ LPAREN! fexpr COMMA! fexpr RPAREN!
    | "int"^ LPAREN! fexpr RPAREN!
    | "frac"^ LPAREN! fexpr RPAREN!
    | "sgn"^ LPAREN! fexpr RPAREN!
    | "norm"^ LPAREN! fexpr COMMA! fexpr RPAREN!
    | "dist"^ LPAREN! ref COMMA! ref RPAREN!
    | "t"^
    | "dt"^
    | ID
    ;

float_
    : FLOAT
    | INT
    ;















/*  **************************************************************
    LE LEXEUR
    definition du nom des attributs, des symboles, des synonymes,
    et de la traduction des mots dans une autre langue.
*/






class CalcLexer extends Lexer;

options {
    testLiterals = false;
    charVocabulary = '\3'..'\377';
    defaultErrorHandler=false;
    k=17; // 6
}

COMMENT
  :  "//" (~('\n'|'\r'))*
    { _ttype = ANTLR_USE_NAMESPACE(antlr)Token::SKIP; }
  ;

protected
DIGITS
	:	( '0'..'9' )+
	;

INT
: ('.')=> '.' DIGITS
        { $setType(FLOAT); }
| (DIGITS '.')=> DIGITS '.' ( '0'..'9' )*
        { $setType(FLOAT); }
| DIGITS
;
/*INT: ((DIGITS)? '.')=> (DIGITS)? '.' DIGITS
        { $setType(FLOAT); }
   | DIGITS
   ;
*/
NEWLINE: '\n';
LPAREN:	 '(';
RPAREN:	 ')';
STAR:	 '*';
DIV:	 '/';
PLUS:	 '+';
MINUS:	 '-';
SEMI:	 ';';
COLON:   ':';
COMMA:   ',';
AT:      '@';
EQUAL:   '=';
GREATER: '>';
LESS:    '<';
GEQ:     ">=";
LEQ:     "<=";
NEQ:     "<>";

protected
LETTER
    : ( 'a'..'z' | 'A'..'Z' | '_' )
    ;

protected
ATTRIBUTES
: ( "dx" | "dy" | 'x' | 'y' | "angle" | "icon" | "mem" | "width" | "height" | "color" | "red" | "green" | "blue" | "gray" | "shockx" | "shocky" | "visible" | "contact" );

protected
FRENCH_VOCAB: "si" | "alors" | "sinon" | "finsi" | "met_en"
    | "icone" | "hauteur" | "largeur" | "couleur" | "chocx" | "chocy"
    | "rouge" | "vert" | "bleu" | "gris"
    | "ROUGE" | "VERT" | "BLEU" | "JAUNE" | "VIOLET" | "CYAN" | "BLANC" | "NOIR"
    | "racine" | "carre" | "norme" | "ent" | "et" | "ou"
    | "souris" | "mulot" | "moi"
    | "plusproche" | "plusproche2" | "plusproche3" | "pp" | "pp2" | "pp3"
    | "plusproche_gauche" | "plusproche_droite" | "plusproche_haut" | "plusproche_bas"  | "ppg" | "ppd" | "pph" | "ppb"
    | "prec" | "suiv" | "VIDE" | "clic"
    | "touche_gauche" | "touche_droite" | "touche_haut" | "touche_bas" | "touche_espace"
    | "plusproche_vers" | "pp_vers"
    ;

protected
SYNONYMS: "grey" | "this" | "NULL" | "atan2" | "nearest2" | "nearest3" | "nearest_indir";

ID  options { testLiterals = true; }
    : (ATTRIBUTES)=> (
	  "dx" { $setType(LITERAL_dx); }
	| "dy" { $setType(LITERAL_dy); }
	| 'x' { $setType(LITERAL_x); }
	| 'y' { $setType(LITERAL_y); }
	| "angle" { $setType(LITERAL_angle); }
	| "icon" { $setType(LITERAL_icon); }
	| "mem" { $setType(LITERAL_mem); }
	| "height" { $setType(LITERAL_height); }
	| "width" { $setType(LITERAL_width); }
	| "color" { $setType(LITERAL_color); }
	| "shockx" { $setType(LITERAL_shockx); }
	| "shocky" { $setType(LITERAL_shocky); }
	| "red" { $setType(LITERAL_red); }
	| "green" { $setType(LITERAL_green); }
	| "blue" { $setType(LITERAL_blue); }
	| "gray" { $setType(LITERAL_gray); }
	| "visible" { $setType(LITERAL_visible); }
	| "contact" { $setType(LITERAL_contact); }
    )
    // SYNONYMS
    | (SYNONYMS)=>  (
	  "grey"  { $setType(LITERAL_gray);  }
        | "this"  { $setType(LITERAL_me);    }
        | "NULL"  { $setType(LITERAL_EMPTY); }
	| "atan2" { $setType(LITERAL_ang);   }
	| "nearest2" { $setType(LITERAL_nearest_second); }
	| "nearest3" { $setType(LITERAL_nearest_third);  }
	| "nearest_indir" { $setType(LITERAL_nearest_in_dir);  }
    )
    // TRADUCTIONS
    | (FRENCH_VOCAB)=>  (
	  "si" { $setType(LITERAL_if); }
	| "alors" { $setType(LITERAL_then); }
	| "sinon" { $setType(LITERAL_else); }
	| "finsi" { $setType(LITERAL_endif); }
	| "met_en" { $setType(LITERAL_move_to); }
	| "icone" { $setType(LITERAL_icon); }
	| "hauteur" { $setType(LITERAL_height); }
	| "largeur" { $setType(LITERAL_width); }
	| "couleur" { $setType(LITERAL_color); }
	| "chocx" { $setType(LITERAL_shockx); }
	| "chocy" { $setType(LITERAL_shocky); }
	| "rouge" { $setType(LITERAL_red); }
	| "vert" { $setType(LITERAL_green); }
	| "bleu" { $setType(LITERAL_blue); }
	| "gris" { $setType(LITERAL_gray); }
	| "ROUGE" { $setType(LITERAL_RED); }
	| "VERT" { $setType(LITERAL_GREEN); }
	| "BLEU" { $setType(LITERAL_BLUE); }
	| "JAUNE" { $setType(LITERAL_YELLOW); }
	| "VIOLET" { $setType(LITERAL_PURPLE); }
	| "CYAN" { $setType(LITERAL_CYAN); }
	| "BLANC" { $setType(LITERAL_WHITE); }
	| "NOIR" { $setType(LITERAL_BLACK); }
	| "racine" { $setType(LITERAL_sqrt); }
	| "carre" { $setType(LITERAL_sqr); }
	| "norme" { $setType(LITERAL_norm); }
	| "ent" { $setType(LITERAL_int); }
	| "et" { $setType(LITERAL_and); }
	| "ou" { $setType(LITERAL_or); }
	| "souris" { $setType(LITERAL_mouse); }
	| "mulot" { $setType(LITERAL_mouse); }
	| "moi" { $setType(LITERAL_me); }
	| "plusproche" { $setType(LITERAL_nearest); }
	| "plusproche2" { $setType(LITERAL_nearest_second); }
	| "plusproche3" { $setType(LITERAL_nearest_third);  }
	| "pp" { $setType(LITERAL_nearest); }
	| "pp2" { $setType(LITERAL_nearest_second); }
	| "pp3" { $setType(LITERAL_nearest_third);  }
	| "plusproche_gauche" { $setType(LITERAL_nearest_left);  }
	| "plusproche_droite" { $setType(LITERAL_nearest_right); }
	| "plusproche_haut"   { $setType(LITERAL_nearest_up);    }
	| "plusproche_bas"    { $setType(LITERAL_nearest_down);  }
	| "ppg"   { $setType(LITERAL_nearest_left);  }
	| "ppd"   { $setType(LITERAL_nearest_right); }
	| "pph"   { $setType(LITERAL_nearest_up);    }
	| "ppb"   { $setType(LITERAL_nearest_down);  }
	| "prec" { $setType(LITERAL_prev); }
	| "suiv" { $setType(LITERAL_next); }
	| "VIDE" { $setType(LITERAL_EMPTY); }
	| "clic" { $setType(LITERAL_click); }
	| "touche_gauche" { $setType(LITERAL_key_left);   }
	| "touche_droite" { $setType(LITERAL_key_right);  }
	| "touche_haut"   { $setType(LITERAL_key_up);     }
	| "touche_bas"    { $setType(LITERAL_key_down);   }
	| "touche_espace" { $setType(LITERAL_key_space);  }
	| "plusproche_vers" { $setType(LITERAL_nearest_in_dir);  }
	| "pp_vers" { $setType(LITERAL_nearest_in_dir);  }
    )
    | LETTER ( LETTER | '0'..'9' )*
;


WS_ :   (   ' '
        |   '\t'
        |   '\r'
        )
        { _ttype = ANTLR_USE_NAMESPACE(antlr)Token::SKIP; } //ignore this token
    ;

